import importlib.util

from .context_algorithm import ContextAlgorithm, ContextRule, ContextRuleDirection
from .qualifier import (
    ATTR_QUALIFIERS,
    Qualifier,
    QualifierDetector,
    QualifierFactory,
    get_qualifiers,
    set_qualifiers,
)

if importlib.util.find_spec("transformers") is not None:
    from .transformer import (
        ExperiencerTransformer,
        NegationTransformer,
        QualifierTransformer,
    )



import importlib.resources
import itertools
import json
import re
import warnings
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import Iterator, Optional, Union

import intervaltree as ivt
from spacy.language import Language
from spacy.matcher import Matcher, PhraseMatcher
from spacy.tokens import Doc, Span

from clinlp.qualifier.qualifier import (
    ATTR_QUALIFIERS,
    Qualifier,
    QualifierDetector,
    QualifierFactory,
)
from clinlp.util import clinlp_autocomponent


class ContextRuleDirection(Enum):
    """
    Direction of a Context rule, as in the original Context Algorithm.
    """

    PRECEDING = 1
    FOLLOWING = 2
    PSEUDO = 3
    TERMINATION = 4


@dataclass
class ContextRule:
    """
    A Context rule, as in the original Context Algorithm.

    Args:
        pattern: The pattern to look for in text. Either a string, or a spacy pattern
        (list).
        qualifier: The qualifier to apply.
        direction: The Context rule direction.
        max_scope: The maximum scope (number of tokens) of the trigger, or None for
        using sentence boundaries.
    """

    pattern: Union[str, list[dict[str, str]]]
    qualifier: Qualifier
    direction: ContextRuleDirection
    max_scope: Optional[int] = None


class _MatchedContextPattern:
    """
    A matched Context pattern, that should be processed further.
    """

    def __init__(
        self, rule: ContextRule, start: int, end: int, offset: int = 0
    ) -> None:
        self.rule = rule
        self.start = start + offset
        self.end = end + offset
        self.scope = None

    def initialize_scope(self, sentence: Span) -> None:
        """
        Sets the scope this pattern ranges over, based on the sentence. This is either
        the window determined in the `max_scope` of the rule, or the sentence
        boundaries if no `max_scope` is set.
        """

        max_scope = self.rule.max_scope or len(sentence)

        if max_scope < 1:
            raise ValueError(f"max_scope must be at least 1, but got {max_scope}")

        if self.rule.direction == ContextRuleDirection.PRECEDING:
            self.scope = (self.start, min(self.end + max_scope, sentence.end))

        elif self.rule.direction == ContextRuleDirection.FOLLOWING:
            self.scope = (max(self.start - max_scope, sentence.start), self.end)


_defaults_context_algorithm = {
    "phrase_matcher_attr": "TEXT",
    "load_rules": True,
    "rules": str(
        importlib.resources.files("clinlp.resources").joinpath("context_rules.json")
    ),
}


@Language.factory(
    name="clinlp_context_algorithm",
    requires=["doc.sents", "doc.ents"],
    assigns=[f"span._.{ATTR_QUALIFIERS}"],
    default_config=_defaults_context_algorithm,
)
@clinlp_autocomponent
class ContextAlgorithm(QualifierDetector):
    """
    Implements the Context algorithm (https://doi.org/10.1016%2Fj.jbi.2009.05.002) as
    a spaCy pipeline component.

    Args:
        nlp: The Spacy language object to use
        phrase_matcher_attr: The token attribute to match phrases on (e.g. TEXT, ORTH,
        NORM).
        load_rules: Whether to parse any rules. Set this to `False` to use
        ContextAlgorithm.add_rules to add ContextRules manually.
        rules: A dictionary of rules, or a path to a json containing the rules
        (see clinlp.resources dir for example).
    """

    def __init__(
        self,
        nlp: Language,
        phrase_matcher_attr: str = _defaults_context_algorithm["phrase_matcher_attr"],
        load_rules=_defaults_context_algorithm["load_rules"],
        rules: Optional[Union[str | dict]] = _defaults_context_algorithm["rules"],
    ) -> None:
        self._nlp = nlp

        self._matcher = Matcher(self._nlp.vocab)
        self._phrase_matcher = PhraseMatcher(self._nlp.vocab, attr=phrase_matcher_attr)

        self.rules = {}
        self._qualifier_factories = {}

        if load_rules:
            if rules is None:
                raise ValueError(
                    "Did not provide rules. Set `load_rules` to False if you "
                    "want to add `ContextRule` manually."
                )

            rules = self._parse_rules(rules)
            self.add_rules(rules)

    @property
    def qualifier_factories(self) -> dict[str, QualifierFactory]:
        return self._qualifier_factories

    def add_rule(self, rule: ContextRule) -> None:
        """
        Add a rule.
        """
        rule_key = f"rule_{len(self.rules)}"
        self.rules[rule_key] = rule

        if isinstance(rule.pattern, str):
            self._phrase_matcher.add(key=rule_key, docs=[self._nlp(rule.pattern)])

        elif isinstance(rule.pattern, list):
            self._matcher.add(key=rule_key, patterns=[rule.pattern])

        else:
            raise ValueError(
                f"Don't know how to process ContextRule with pattern of "
                f"type {type(rule.pattern)}"
            )

    def add_rules(self, rules: list[ContextRule]) -> None:
        """
        Add multiple rules.
        """
        for rule in rules:
            self.add_rule(rule)

    @staticmethod
    def _parse_qualifier(
        qualifier: str, qualifier_factories: dict[str, QualifierFactory]
    ) -> Qualifier:
        """
        Parse a Qualifier from string.

        Args:
            qualifier: The qualifier (e.g. Negation.NEGATED).

        Returns: A qualifier, as specified.
        """

        match_regexp = r"\w+\.\w+"

        if not re.match(match_regexp, qualifier):
            raise ValueError(
                f"Cannot parse qualifier {qualifier}, please adhere to format "
                f"{match_regexp} (e.g. NegationQualifier.NEGATED)"
            )

        qualifier_class, qualifier = qualifier.split(".")

        return qualifier_factories[qualifier_class].create(value=qualifier)

    @staticmethod
    def _parse_direction(direction: str) -> ContextRuleDirection:
        """
        Parse a Context direction.

        Args:
            direction: The direction (e.g. preceding, following, etc.).

        Returns: THe ContextRuleDirection.
        """
        return ContextRuleDirection[direction.upper()]

    def _parse_rules(self, rules: Union[str | dict]) -> list[ContextRule]:
        if isinstance(rules, str):
            with open(rules, "rb") as file:
                rules = json.load(file)

        for qualifier in rules["qualifiers"]:
            self._qualifier_factories[qualifier["name"]] = QualifierFactory(**qualifier)

        qualifier_rules = []

        for rule in rules["rules"]:
            qualifier = self._parse_qualifier(
                rule["qualifier"], self.qualifier_factories
            )
            direction = self._parse_direction(rule["direction"])
            max_scope = rule.get("max_scope", None)

            qualifier_rules += [
                ContextRule(pattern, qualifier, direction, max_scope)
                for pattern in rule["patterns"]
            ]

        return qualifier_rules

    @staticmethod
    def _get_sentences_having_entity(doc: Doc) -> Iterator[Span]:
        """
        Return sentences in a doc that have at least one entity.
        """
        return (sent for sent in doc.sents if len(sent.ents) > 0)

    def _get_rule_from_match_id(self, match_id: int) -> ContextRule:
        """
        Get the rule that was matched, from the match_id (first element of match tuple
        returned by matcher).
        """
        return self.rules[self._nlp.vocab.strings[match_id]]

    @staticmethod
    def _group_matched_patterns(
        matched_patterns: list[_MatchedContextPattern],
    ) -> defaultdict:
        """
        Group matched patterns by qualifier and direction.
        """
        groups = defaultdict(lambda: defaultdict(list))

        for matched_rule in matched_patterns:
            groups[matched_rule.rule.qualifier][matched_rule.rule.direction].append(
                matched_rule
            )

        return groups

    @staticmethod
    def _limit_scopes_from_terminations(
        scopes: ivt.IntervalTree, terminations: list[_MatchedContextPattern]
    ) -> ivt.IntervalTree:
        """
        Determine the scope of terminating matched context pattern, return them
        as IntervalTree.
        """

        for terminate_match in terminations:
            for interval in scopes.overlap(terminate_match.start, terminate_match.end):
                scopes.remove(interval)
                match = interval.data

                if (
                    match.rule.direction == ContextRuleDirection.PRECEDING
                    and terminate_match.start >= match.end
                ):
                    match.scope = (match.scope[0], terminate_match.start)

                if (
                    match.rule.direction == ContextRuleDirection.FOLLOWING
                    and terminate_match.end <= match.start
                ):
                    match.scope = (terminate_match.end, match.scope[1])

                scopes[match.scope[0] : match.scope[1]] = match

        return scopes

    def _compute_match_scopes(
        self, matched_patterns: list[_MatchedContextPattern]
    ) -> ivt.IntervalTree:
        """
        Compute the scope for each matched pattern, return them as an IntervalTree.
        """
        match_scopes = ivt.IntervalTree()

        for qualifier_matches in self._group_matched_patterns(
            matched_patterns
        ).values():
            preceding = qualifier_matches[ContextRuleDirection.PRECEDING]
            following = qualifier_matches[ContextRuleDirection.FOLLOWING]
            pseudo = qualifier_matches[ContextRuleDirection.PSEUDO]
            termination = qualifier_matches[ContextRuleDirection.TERMINATION]

            qualifier_matches = ivt.IntervalTree()

            # Following, preceding
            for match in preceding + following:
                qualifier_matches[match.start : match.end] = match

            # Pseudo
            for match in pseudo:
                qualifier_matches.remove_overlap(match.start, match.end)

            # Termination
            qualifier_scopes = ivt.IntervalTree(
                ivt.Interval(i.data.scope[0], i.data.scope[1], i.data)
                for i in qualifier_matches
            )

            match_scopes |= self._limit_scopes_from_terminations(
                qualifier_scopes, termination
            )

        return match_scopes

    def _detect_qualifiers(self, doc: Doc):
        """
        Apply the Context Algorithm to a doc.
        """

        if len(self.rules) == 0:
            raise RuntimeError("Cannot match qualifiers without any ContextRule.")

        for sentence in self._get_sentences_having_entity(doc):
            with warnings.catch_warnings():
                # a UserWarning will trigger when one of the matchers is empty
                warnings.simplefilter("ignore", UserWarning)

                matches = itertools.chain(
                    self._matcher(sentence), self._phrase_matcher(sentence)
                )

            matched_patterns = []

            for match_id, start, end in matches:
                rule = self._get_rule_from_match_id(match_id)

                # spacy Matcher handles offset differently than PhraseMatcher,
                # when applying the matcher to a sentence
                offset = sentence.start if isinstance(rule.pattern, list) else 0

                pattern = _MatchedContextPattern(
                    rule=self._get_rule_from_match_id(match_id),
                    start=start,
                    end=end,
                    offset=offset,
                )

                pattern.initialize_scope(sentence)
                matched_patterns.append(pattern)

            match_scopes = self._compute_match_scopes(matched_patterns)

            for ent in sentence.ents:
                for match_interval in match_scopes.overlap(ent.start, ent.end):
                    if (ent.start + 1 > match_interval.data.end) or (
                        ent.end < match_interval.data.start + 1
                    ):
                        self.add_qualifier_to_ent(
                            ent, match_interval.data.rule.qualifier
                        )

    def __len__(self) -> int:
        return len(self.rules)



import importlib.util

from .entity import EntityMatcher, Term
from .language import Clinlp
from .normalizer import Normalizer
from .qualifier import ContextAlgorithm
from .sentencizer import Sentencizer

if importlib.util.find_spec("transformers") is not None:
    from .qualifier import NegationTransformer



from typing import Optional

import intervaltree as ivt
import numpy as np
import pandas as pd
import pydantic
from spacy.language import Doc, Language
from spacy.matcher import Matcher, PhraseMatcher
from spacy.tokens import Span

from clinlp.util import clinlp_autocomponent

_defaults_clinlp_ner = {
    "attr": "TEXT",
    "proximity": 0,
    "fuzzy": 0,
    "fuzzy_min_len": 0,
    "pseudo": False,
}
_non_phrase_matcher_fields = ["proximity", "fuzzy", "fuzzy_min_len"]


class Term(pydantic.BaseModel):
    phrase: str
    attr: Optional[str] = None
    proximity: Optional[int] = None
    fuzzy: Optional[int] = None
    fuzzy_min_len: Optional[int] = None
    pseudo: Optional[bool] = None

    model_config = {"extra": "ignore"}

    def __init__(self, phrase: str, **kwargs):
        """This init makes sure Term accepts phrase as a positional argument,
        which is more readable in large concept lists."""
        super().__init__(phrase=phrase, **kwargs)

    def to_spacy_pattern(self, nlp: Language):
        fields = {
            field: getattr(self, field) or _defaults_clinlp_ner[field]
            for field in ["attr", "proximity", "fuzzy", "fuzzy_min_len", "pseudo"]
        }

        spacy_pattern = []

        phrase_tokens = [token.text for token in nlp.tokenizer(self.phrase)]

        for i, token in enumerate(phrase_tokens):
            if (fields["fuzzy"] > 0) and (len(token) >= fields["fuzzy_min_len"]):
                token_pattern = {f"FUZZY{fields['fuzzy']}": token}
            else:
                token_pattern = token

            spacy_pattern.append({fields["attr"]: token_pattern})

            if i != len(phrase_tokens) - 1:
                for _ in range(fields["proximity"]):
                    spacy_pattern.append({"OP": "?"})

        return spacy_pattern


def create_concept_dict(path: str, concept_col: str = "concept") -> dict:
    """Transforms source concept data to a dictionary that the clinlp
    entity matcher can read. Takes the path to a csv file where each
    row is a distinct word or sentence (term) that belongs to a concept.
    """

    df = pd.read_csv(path).replace([np.nan], [None])

    try:
        df["term"] = df.apply(lambda x: Term(**x.to_dict()), axis=1)
    except pydantic.ValidationError as e:
        msg = (
            "There is a value in your input csv which cannot be"
            "parsed. Please refer to the above error for more details."
        )

        raise RuntimeError(msg) from e

    df = df.groupby(concept_col)["term"].apply(list).reset_index()

    return dict(zip(df["concept"], df["term"]))


@Language.factory(
    name="clinlp_entity_matcher",
    requires=["doc.sents", "doc.ents"],
    assigns=["doc.ents"],
    default_config=_defaults_clinlp_ner,
)
@clinlp_autocomponent
class EntityMatcher:
    def __init__(
        self,
        nlp: Language,
        attr: Optional[str] = _defaults_clinlp_ner["attr"],
        proximity: Optional[int] = _defaults_clinlp_ner["proximity"],
        fuzzy: Optional[int] = _defaults_clinlp_ner["fuzzy"],
        fuzzy_min_len: Optional[int] = _defaults_clinlp_ner["fuzzy_min_len"],
        pseudo: Optional[bool] = _defaults_clinlp_ner["pseudo"],
    ):
        self.nlp = nlp
        self.attr = attr

        self.term_args = {
            "attr": attr,
            "proximity": proximity,
            "fuzzy": fuzzy,
            "fuzzy_min_len": fuzzy_min_len,
            "pseudo": pseudo,
        }

        self._matcher = Matcher(self.nlp.vocab)
        self._phrase_matcher = PhraseMatcher(self.nlp.vocab, attr=self.attr)

        self._terms = {}
        self._concepts = {}

    @property
    def _use_phrase_matcher(self):
        return all(
            self.term_args[field] == _defaults_clinlp_ner[field]
            for field in _non_phrase_matcher_fields
            if field in self.term_args
        )

    def load_concepts(self, concepts: str | dict):
        for concept, concept_terms in concepts.items():
            for concept_term in concept_terms:
                identifier = str(len(self._terms))

                self._terms[identifier] = concept_term
                self._concepts[identifier] = concept

                if isinstance(concept_term, str):
                    if self._use_phrase_matcher:
                        self._phrase_matcher.add(
                            key=identifier, docs=[self.nlp(concept_term)]
                        )
                    else:
                        concept_term = Term(
                            phrase=concept_term, **self.term_args
                        ).to_spacy_pattern(self.nlp)
                        self._matcher.add(key=identifier, patterns=[concept_term])

                elif isinstance(concept_term, list):
                    self._matcher.add(key=identifier, patterns=[concept_term])

                elif isinstance(concept_term, Term):
                    term_args_with_override = {}

                    for field, value in self.term_args.items():
                        if getattr(concept_term, field) is not None:
                            term_args_with_override[field] = getattr(
                                concept_term, field
                            )
                        else:
                            term_args_with_override[field] = value

                    self._matcher.add(
                        key=identifier,
                        patterns=[
                            Term(
                                phrase=concept_term.phrase, **term_args_with_override
                            ).to_spacy_pattern(self.nlp)
                        ],
                    )

                else:
                    raise ValueError(
                        f"Not sure how to load a term with type {type(concept_term)}, "
                        f"please provide str, list or clinlp.Term"
                    )

    def _get_matches(self, doc: Doc):
        if len(self._terms) == 0:
            raise RuntimeError("No concepts added.")

        matches = []

        if len(self._matcher) > 0:
            matches += list(self._matcher(doc))

        if len(self._phrase_matcher) > 0:
            matches += list(self._phrase_matcher(doc))

        return matches

    @staticmethod
    def _resolve_ents_overlap(ents: list[Span]) -> list[Span]:
        """
        Resolves overlap between spans. Current logic: take the longest.

        Args:
            ents: The input Spans, with possible overlap.

        Returns: The Spans without any overlap.
        """

        if len(ents) == 0:
            return ents

        ents = sorted(ents, key=lambda span: span.start)

        disjoint_ents = [ents[0]]

        for _, ent in enumerate(ents[1:]):
            if ent.start < disjoint_ents[-1].end:
                if len(str(disjoint_ents[-1])) < len(str(ent)):
                    disjoint_ents[-1] = ent
            else:
                disjoint_ents.append(ent)

        return disjoint_ents

    def __call__(self, doc: Doc):
        matches = self._get_matches(doc)

        pos_matches = []
        neg_matches = ivt.IntervalTree()

        for match_id, start, end in matches:
            rule_id = self.nlp.vocab.strings[match_id]
            term = self._terms[rule_id]

            if isinstance(term, Term) and term.pseudo:
                neg_matches[start:end] = rule_id
            else:
                pos_matches.append((rule_id, start, end))

        ents = []

        for match_id, start, end in pos_matches:
            if not any(
                self._concepts[match_id] == self._concepts[neg_match_id.data]
                for neg_match_id in neg_matches.overlap(start, end)
            ):
                ents.append(
                    Span(doc=doc, start=start, end=end, label=self._concepts[match_id])
                )

        ents = self._resolve_ents_overlap(ents)
        doc.set_ents(entities=ents)

        return doc



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional

from spacy.tokens import Doc, Span

ATTR_QUALIFIERS = "qualifiers"
ATTR_QUALIFIERS_STR = f"{ATTR_QUALIFIERS}_str"
ATTR_QUALIFIERS_DICT = f"{ATTR_QUALIFIERS}_dict"


def qualifiers_to_str(ent: Span) -> Optional[set[str]]:
    qualifiers = getattr(getattr(ent, "_"), ATTR_QUALIFIERS)

    if qualifiers is None:
        return None

    return {str(q) for q in qualifiers}


def qualifiers_to_dict(ent: Span) -> Optional[list[dict]]:
    qualifiers = getattr(getattr(ent, "_"), ATTR_QUALIFIERS)

    if qualifiers is None:
        return None

    return [q.to_dict() for q in qualifiers]


Span.set_extension(name=ATTR_QUALIFIERS, default=None)
Span.set_extension(name=ATTR_QUALIFIERS_STR, getter=qualifiers_to_str)
Span.set_extension(name=ATTR_QUALIFIERS_DICT, getter=qualifiers_to_dict)


def get_qualifiers(entity: Span) -> set["Qualifier"]:
    return getattr(getattr(entity, "_"), ATTR_QUALIFIERS)


def set_qualifiers(entity: Span, qualifiers: set["Qualifier"]) -> None:
    setattr(getattr(entity, "_"), ATTR_QUALIFIERS, qualifiers)


@dataclass(frozen=True)
class Qualifier:
    name: str = field(compare=True)
    value: str = field(compare=True)
    is_default: bool = field(compare=True)
    prob: Optional[float] = field(default=None, compare=False)

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "value": self.value,
            "is_default": self.is_default,
            "prob": self.prob,
        }

    def __str__(self) -> str:
        return f"{self.name}.{self.value}"


class QualifierFactory:
    def __init__(self, name: str, values: list[str], default: Optional[str] = None):
        self.name = name
        self.default = default or values[0]

        if len(set(values)) != len(values):
            raise ValueError(f"Please do not provide any duplicate values ({values})")

        if self.default not in values:
            raise ValueError(f"Default {default} not in provided value {values}")

        self.values = values

    def create(self, value: Optional[str] = None, **kwargs) -> Qualifier:
        if value is None:
            value = self.default

        if value not in self.values:
            raise ValueError(
                f"The qualifier {self.name} cannot take value '{value}'. "
                f"Please choose one of {self.values}."
            )

        is_default = value == self.default

        return Qualifier(name=self.name, value=value, is_default=is_default, **kwargs)


class QualifierDetector(ABC):
    """For usage as a spaCy pipeline component"""

    @property
    @abstractmethod
    def qualifier_factories(self) -> dict[str, QualifierFactory]:
        pass

    @abstractmethod
    def _detect_qualifiers(self, doc: Doc) -> None:
        pass

    @staticmethod
    def add_qualifier_to_ent(entity: Span, new_qualifier: Qualifier) -> None:
        qualifiers = get_qualifiers(entity)

        if qualifiers is None:
            raise RuntimeError(
                "Cannot add qualifier to entity with non-initialized qualifiers."
            )

        try:
            old_qualifier = next(
                iter(q for q in qualifiers if q.name == new_qualifier.name)
            )

            qualifiers.remove(old_qualifier)
            qualifiers.add(new_qualifier)

        except StopIteration:
            qualifiers.add(new_qualifier)

        set_qualifiers(entity, qualifiers)

    def _initialize_ent_qualifiers(self, entity: Span) -> None:
        if get_qualifiers(entity) is None:
            set_qualifiers(entity, set())

        for _, factory in self.qualifier_factories.items():
            self.add_qualifier_to_ent(entity, factory.create())

    def __call__(self, doc: Doc) -> Doc:
        if len(doc.ents) == 0:
            return doc

        for ent in doc.ents:
            self._initialize_ent_qualifiers(ent)

        self._detect_qualifiers(doc)

        return doc



from typing import Optional

import spacy.tokens
from spacy.language import Language

from clinlp.util import clinlp_autocomponent

_defaults_sentencizer = {
    "sent_end_chars": [".", "!", "?", "\n", "\r"],
    "sent_start_punct": ["-", "*", "[", "("],
}


@Language.factory(
    "clinlp_sentencizer",
    assigns=["token.is_sent_start", "doc.sents"],
    default_config=_defaults_sentencizer,
)
@clinlp_autocomponent
class Sentencizer:
    def __init__(
        self,
        sent_end_chars: Optional[list[str]] = None,
        sent_start_punct: Optional[list[str]] = None,
    ):
        self.sent_end_chars = (
            _defaults_sentencizer["sent_end_chars"]
            if sent_end_chars is None
            else sent_end_chars
        )
        self.sent_start_punct = (
            _defaults_sentencizer["sent_start_punct"]
            if sent_start_punct is None
            else sent_start_punct
        )

        self.sent_end_chars = set(self.sent_end_chars)
        self.sent_start_punct = set(self.sent_start_punct)

    def _token_can_start_sent(self, token: spacy.tokens.Token) -> bool:
        """
        Determines whether a token can start a sentence
        """
        return (
            token.text[0].isalnum()
            or (token.text[0] in {"["})
            or (token.text in self.sent_start_punct)
        )

    def _token_can_end_sent(self, token: spacy.tokens.Token):
        """
        Determines whether a token can end a sentence
        """
        return token.text in self.sent_end_chars

    def _get_sentence_starts(self, doc: spacy.tokens.Doc) -> list[bool]:
        if len(doc) == 0:
            return []

        sentence_starts = [False] * len(doc)

        if self._token_can_start_sent(doc[0]):
            sentence_starts[0] = True

        seen_end_char = True

        for i, token in enumerate(doc):
            if seen_end_char and self._token_can_start_sent(token):
                sentence_starts[i] = True
                seen_end_char = False

            if self._token_can_end_sent(token):
                seen_end_char = True

        return sentence_starts

    def __call__(self, doc: spacy.tokens.Doc):
        if len(doc) == 0:
            return doc

        sentence_starts = self._get_sentence_starts(doc)

        for is_sent_start, token in zip(sentence_starts, doc):
            token.is_sent_start = is_sent_start

        return doc



import unicodedata

from spacy import Language
from spacy.tokens import Doc

from clinlp.util import clinlp_autocomponent

_defaults_normalizer = {"lowercase": True, "map_non_ascii": True}


@Language.factory(
    "clinlp_normalizer", assigns=["token.norm"], default_config=_defaults_normalizer
)
@clinlp_autocomponent
class Normalizer:
    def __init__(
        self,
        lowercase=_defaults_normalizer["lowercase"],
        map_non_ascii=_defaults_normalizer["map_non_ascii"],
    ):
        self.lowercase = lowercase
        self.map_non_ascii = map_non_ascii

    @staticmethod
    def _lowercase(text: str) -> str:
        return text.lower()

    @staticmethod
    def _map_non_ascii_char(char: str) -> str:
        if len(char) != 1:
            raise ValueError(
                "Please only use the _map_non_ascii_char method on strings of length 1."
            )

        normalized_char = unicodedata.normalize("NFD", char)
        normalized_char = str(normalized_char.encode("ascii", "ignore").decode("utf-8"))

        return normalized_char if len(normalized_char) > 0 else char

    def _map_non_ascii_string(self, text: str) -> str:
        return "".join(self._map_non_ascii_char(char) for char in text)

    def __call__(self, doc: Doc):
        if len(doc) == 0:
            return doc

        for token in doc:
            normalized_text = token.text

            if self.lowercase:
                normalized_text = self._lowercase(normalized_text)

            if self.map_non_ascii:
                normalized_text = self._map_non_ascii_string(normalized_text)

            token.norm_ = normalized_text

        return doc



import importlib.metadata
import warnings
from typing import Callable, Optional

import spacy.lang.char_classes
import spacy.lang.nl.tokenizer_exceptions
import spacy.lang.punctuation
import spacy.lang.tokenizer_exceptions
from spacy.language import BaseDefaults, Language
from spacy.symbols import ORTH
from spacy.util import update_exc

from clinlp.exceptions import VersionMismatchWarning

CLINLP_ABBREVIATIONS = [
    "dhr.",
    "dr.",
    "dh.",
    "drs.",
    "ds.",
    "gg.",
    "ggn.",
    "grg.",
    "ing.",
    "ir.",
    "medepat.",
    "mej.",
    "mevr.",
    "mr.",
    "mw.",
    "prof.",
    "pt.",
    "1.",
    "2.",
    "3.",
    "4.",
    "5.",
    "6.",
    "7.",
    "8.",
    "9.",
    "10.",
    "11.",
    "12.",
    "I.",
    "II.",
    "III.",
    "IV.",
    "V.",
    "VI.",
    "VII.",
    "VIII.",
    "IX.",
    "X.",
    "t.b.v.",
    "a.n.",
    "internat.eenh.",
    "o.a.",
    "oa.",
    "tel.",
    "i.o.m.",
    "z.n.",
    "o.l.v.",
    "z.n.",
    "e.c.i.",
    "s.c.",
    "etc.",
    "g.b.",
    "t.p.v.",
    "vlgs.",
    "o.b.v.",
    "t.h.v.",
    "i.o.m.",
    "i.v.",
    "internat.eenh.",
    "m.i.",
    "v.v.",
    "i.p.",
    "i.v.m.",
    "i.c.m.",
    "m.b.v.",
    "wv.",
    "pol.",
    "tr.",
    "vv.",
    "mi.",
]

CLINLP_REMOVE_ABBREVIATIONS = [
    "ts.",
]

CLINLP_UNITS = [
    "cc",
    "CC",
    "l",
    "L",
    "ml",
    "mL",
    "µl",
    "µL",
    "dag",
    "h",
    "hh",
    "hr",
    "j",
    "jr",
    "s",
    "sec",
    "W",
    "w",
    "week",
    "wkn",
    "wk",
    "min",
    "mnd",
    "ms",
    "pm",
    "g",
    "gr",
    "mg",
    "mG",
    "µg",
    "ug",
    "mcg",
    "do",
    "dosis",
    "stuk",
    "e",
    "x",
    "Gy",
    "eh",
    "ie",
    "IE",
    "Ie",
    "IE",
    "mol",
    "mmol",
    "umol",
    "µmol",
    "pmol",
    "u",
    "kcal",
    "kCal",
    "m2",
    "µm",
    "mmhg",
    "mmHg",
    "km",
    "km²",
    "km³",
    "m",
    "m²",
    "m³",
    "dm",
    "dm²",
    "dm³",
    "cm",
    "cm²",
    "cm³",
    "mm",
    "mm²",
    "mm³",
    "ha",
    "µm",
    "nm",
    "yd",
    "in",
    "ft",
    "kg",
    "g",
    "mg",
    "µg",
    "t",
    "lb",
    "oz",
    "m/s",
    "km/h",
    "kmh",
    "mph",
    "hPa",
    "Pa",
    "mbar",
    "mb",
    "MB",
    "kb",
    "KB",
    "gb",
    "GB",
    "tb",
    "TB",
    "T",
    "G",
    "M",
    "K",
    "%",
    "км",
    "км²",
    "км³",
    "м",
    "м²",
    "м³",
    "см",
    "см²",
    "см³",
    "мм",
    "мм²",
    "мм³",
    "нм",
]

CLINLP_TOKENIZER_EXCEPTIONS = {
    "\n": [{ORTH: "\n"}],
    "\r": [{ORTH: "\r"}],
    "\t": [{ORTH: "\t"}],
    "->": [{ORTH: "->"}],
    "-->": [{ORTH: "-->"}],
    "<->": [{ORTH: "<->"}],
    "=>": [{ORTH: "=>"}],
    "==>": [{ORTH: "==>"}],
    "<=>": [{ORTH: "<=>"}],
    "+/-": [{ORTH: "+/-"}],
    "t/m": [{ORTH: "t/m"}],
    "<=": [{ORTH: "<="}],
    ">=": [{ORTH: ">="}],
    "(?)": [{ORTH: "(?)"}],
    "xdd": [{ORTH: "x"}, {ORTH: "dd"}],
}

CLINLP_ABBREV_TRANSFORMS = [
    lambda x: x,
    lambda x: x.upper(),
    lambda x: x.capitalize(),
]

ALPHA_LOWER = "a-z"
ALPHA_UPPER = "A-Z"
ALPHA = "a-zA-Z"


def _get_abbreviations():
    base = set(spacy.lang.nl.tokenizer_exceptions.abbrevs.copy())

    for abbrev in CLINLP_REMOVE_ABBREVIATIONS:
        base.remove(abbrev)

    return list(base) + CLINLP_ABBREVIATIONS


def _get_tokenizer_exceptions(
    abbreviations: list[str],
    abbrev_transforms: list[Callable[[str], str]] = None,
    keep_emoticons: bool = False,
):
    tokenizer_exceptions = spacy.lang.tokenizer_exceptions.BASE_EXCEPTIONS.copy()

    if not keep_emoticons:
        for emoticon in spacy.lang.tokenizer_exceptions.emoticons:
            del tokenizer_exceptions[emoticon]

    for abbrev_transform in abbrev_transforms:
        abbr_update = {
            abbrev_transform(a): [{ORTH: abbrev_transform(a)}] for a in abbreviations
        }
        tokenizer_exceptions = update_exc(tokenizer_exceptions, abbr_update)

    return update_exc(tokenizer_exceptions, CLINLP_TOKENIZER_EXCEPTIONS)


def _get_list(
    base: list[str], add: Optional[list[str]] = None, remove: Optional[list[str]] = None
):
    """
    Create a list, based on a base list, with some additions and removals.
    """
    _lst = base.copy()

    if add is not None:
        for item in add:
            _lst.append(item)

    if remove is not None:
        for item in remove:
            _lst.remove(item)

    return _lst


def _get_ellipses():
    return spacy.lang.punctuation.LIST_ELLIPSES.copy()


def _get_currencies():
    return spacy.lang.punctuation.LIST_CURRENCY.copy()


def _get_units():
    return CLINLP_UNITS.copy()


def _get_tokenizer_prefix_rules():
    return [
        r"\[(?![A-Z]{3,}-)",
        r"\S+(?=\[[A-Z]{3,}-)",
        r"x(?=[0-9]+)",
        r"`(?=[0-9])",
        r"([0-9]{,5}(\.|,))?[0-9]{,4}" + f"(?=({'|'.join(_get_units())}))",
    ]


def _get_tokenizer_prefixes():
    punct = _get_list(
        base=spacy.lang.punctuation.LIST_PUNCT,
        add=[
            ",,",
            "§",
            "%",
            "=",
            "—",
            "–",
            r"\+(?![0-9])",
            "/",
            "-",
            r"\+",
            "~",
            ",",
            r"\s",
        ],
        remove=[r"\["],
    )

    quotes = _get_list(base=spacy.lang.punctuation.LIST_QUOTES, remove=["`"])

    return (
        punct
        + _get_ellipses()
        + quotes
        + _get_currencies()
        + _get_tokenizer_prefix_rules()
    )


def _get_tokenizer_infix_rules(quotes: list[str]):
    return [
        r"(?<=[{al}])\.(?=[{au}])".format(al=ALPHA_LOWER, au=ALPHA_UPPER),
        r"(?<=[{a}])[,!?](?=[{a}])".format(a=ALPHA),
        r'(?<=[{a}"])[:<>=](?=[{a}])'.format(a=ALPHA),
        r"(?<=[{a}]),(?=[{a}])".format(a=ALPHA),
        r"(?<=[{a}])([{q}\)\]\(\[])(?=[{a}])".format(a=ALPHA, q="".join(quotes)),
        r"(?<=[{a}])--(?=[{a}])".format(a=ALPHA),
        r"-(?![0-9]{1,2}\])",
        r"(?<=[0-9])x(?=[0-9])",
        r"(?<=10)E(?=\d)",
        r"(?<=[0-9])(x?d?d)(?=[0-9])",
    ]


def _get_tokenizer_infixes():
    punct = _get_list(
        base=[],
        add=["/", r"\+", "=", ":", "&", ";", r"\*", "<", r"\(", r"\)", r"\s", r"\^"],
    )

    quotes = _get_list(base=spacy.lang.punctuation.LIST_QUOTES, remove=["`", r"\'"])

    infixes = punct + _get_ellipses() + _get_tokenizer_infix_rules(quotes)

    return infixes


def _get_tokenizer_suffix_rules(
    currencies: list[str], units: list[str], punct: list[str], quotes: list[str]
):
    return [
        r"(?<=[0-9])\+",
        r"(?<=°[FfCcKk])\.",
        r"(?<=[0-9])(?:{c})".format(c="|".join(currencies)),
        r"(?<=[0-9{al}{e}{p}(?:{q})])\.".format(
            al=ALPHA_LOWER, e=r"%²\-\+", q="".join(quotes), p="|".join(punct)
        ),
        r"(?<=[{au}][{au}])\.".format(au=ALPHA_UPPER),
        r"(?<=[0-9]\])\S+",
        r"(?<!([A-Z]-\d|-\d\d))\]",  # tricky one
        r"(?<=[0-9])x",
        r"(?<=[0-9])" + f"({'|'.join(units)})",
        r"\s",
        r"(?<=[0-9])d?d",
        r"(?<=[0-9])(e|de|ste)",
    ]


def _get_tokenizer_suffixes():
    punct = _get_list(
        base=spacy.lang.punctuation.LIST_PUNCT,
        add=["/", "-", "=", "%", r"\+", "~", "''", "—", "–"],
        remove=[r"\]"],
    )

    quotes = _get_list(base=spacy.lang.punctuation.LIST_QUOTES)

    return (
        punct
        + quotes
        + _get_ellipses()
        + _get_tokenizer_suffix_rules(
            currencies=_get_currencies(), units=_get_units(), punct=punct, quotes=quotes
        )
    )


class ClinlpDefaults(BaseDefaults):
    tokenizer_exceptions = _get_tokenizer_exceptions(
        abbreviations=_get_abbreviations(), abbrev_transforms=CLINLP_ABBREV_TRANSFORMS
    )

    prefixes = _get_tokenizer_prefixes()
    infixes = _get_tokenizer_infixes()
    suffixes = _get_tokenizer_suffixes()

    lex_attr_getters = {}
    syntax_iterators = {}
    stop_words = []
    url_match = None
    token_match = None

    writing_system = {"direction": "ltr", "has_case": True, "has_letters": True}


@spacy.registry.languages("clinlp")
class Clinlp(Language):
    lang = "clinlp"
    Defaults = ClinlpDefaults

    def __init__(self, *args, **kwargs):
        meta = dict(kwargs.pop("meta", {}))
        clinlp_version = importlib.metadata.version(__package__ or __name__)

        if "clinlp_version" in meta:
            if meta["clinlp_version"] != clinlp_version:
                warnings.warn(
                    f"This spaCy model was built with clinlp version "
                    f"{meta['clinlp_version']}, but you currently have version "
                    f"{clinlp_version} installed, potentially leading to unexpected "
                    f"results.",
                    VersionMismatchWarning,
                )
        else:
            meta["clinlp_version"] = clinlp_version

        super().__init__(*args, meta=meta, **kwargs)



import statistics
from typing import Callable, Optional, Tuple

import torch
from spacy import Language
from spacy.tokens import Doc, Span
from transformers import AutoTokenizer, RobertaForTokenClassification

from clinlp.qualifier.qualifier import (
    ATTR_QUALIFIERS,
    QualifierDetector,
    QualifierFactory,
)
from clinlp.util import clinlp_autocomponent

HF_FROM_PRETRAINED_NEGATION = {
    "pretrained_model_name_or_path": "UMCU/MedRoBERTa.nl_NegationDetection",
    "revision": "83068ba132b6ce38e9f668c1e3ab636f79b774d3",
}

HF_FROM_PRETRAINED_EXPERIENCER = {
    "pretrained_model_name_or_path": "UMCU/MedRoBERTa.nl_Experiencer",
    "revision": "d9318c4b2b0ab0dfe50afedca58319b2369f1a71",
}

_defaults_qualifier_transformer = {
    "token_window": 32,
    "strip_entities": True,
    "placeholder": None,
    "prob_aggregator": statistics.mean,
}

_defaults_negation_transformer = {
    "negation_threshold": 0.5,
}
_defaults_experiencer_transformer = {
    "token_window": 64,
    "other_threshold": 0.5,
}


class QualifierTransformer(QualifierDetector):
    def __init__(
        self,
        token_window: int = _defaults_qualifier_transformer["token_window"],
        strip_entities: bool = _defaults_qualifier_transformer["strip_entities"],
        placeholder: Optional[str] = _defaults_qualifier_transformer["placeholder"],
        prob_aggregator: int = _defaults_qualifier_transformer["prob_aggregator"],
    ):
        self.token_window = token_window
        self.strip_entities = strip_entities
        self.placeholder = placeholder
        self.prob_aggregator = prob_aggregator

    @staticmethod
    def _get_ent_window(ent: Span, token_window: int) -> Tuple[str, int, int]:
        start_token_i = max(0, ent.start - token_window)
        end_token_i = min(len(ent.doc), ent.end + token_window)

        text_span = ent.doc[start_token_i:end_token_i]

        ent_start_char = ent.start_char - text_span.start_char
        ent_end_char = ent.end_char - text_span.start_char

        return str(text_span), ent_start_char, ent_end_char

    @staticmethod
    def _trim_ent_boundaries(
        text: str, ent_start_char: int, ent_end_char: int
    ) -> Tuple[str, int, int]:
        entity = text[ent_start_char:ent_end_char]

        ent_start_char += len(entity) - len(entity.lstrip())
        ent_end_char -= len(entity) - len(entity.rstrip())

        return text, ent_start_char, ent_end_char

    @staticmethod
    def _fill_ent_placeholder(
        text: str, ent_start_char: int, ent_end_char: int, placeholder: str
    ) -> Tuple[str, int, int]:
        text = text[0:ent_start_char] + placeholder + text[ent_end_char:]
        ent_end_char = ent_start_char + len(placeholder)

        return text, ent_start_char, ent_end_char

    def _prepare_ent(self, ent: Span) -> Tuple[str, int, int]:
        text, ent_start_char, ent_end_char = self._get_ent_window(
            ent, token_window=self.token_window
        )

        if self.strip_entities:
            text, ent_start_char, ent_end_char = self._trim_ent_boundaries(
                text, ent_start_char, ent_end_char
            )

        if self.placeholder is not None:
            text, ent_start_char, ent_end_char = self._fill_ent_placeholder(
                text, ent_start_char, ent_end_char, placeholder=self.placeholder
            )

        return text, ent_start_char, ent_end_char

    def _predict(
        self,
        text: str,
        ent_start_char: int,
        ent_end_char: int,
        prob_indices: list,
        prob_aggregator: Callable,
    ) -> float:
        inputs = self.tokenizer(text, return_tensors="pt")
        output = self.model.forward(inputs["input_ids"])
        probs = torch.nn.functional.softmax(output.logits[0], dim=1).detach().numpy()

        start_token = inputs.char_to_token(ent_start_char)
        end_token = inputs.char_to_token(ent_end_char - 1)

        return prob_aggregator(
            sum(pos[prob_indices]) for pos in probs[start_token : end_token + 1]
        )


@Language.factory(
    name="clinlp_negation_transformer",
    requires=["doc.ents"],
    assigns=[f"span._.{ATTR_QUALIFIERS}"],
    default_config=_defaults_negation_transformer,
)
@clinlp_autocomponent
class NegationTransformer(QualifierTransformer):
    def __init__(
        self,
        nlp: Language,
        negation_threshold: float = _defaults_negation_transformer[
            "negation_threshold"
        ],
        **kwargs,
    ) -> None:
        self.nlp = nlp
        self.negation_threshold = negation_threshold

        self.tokenizer = AutoTokenizer.from_pretrained(**HF_FROM_PRETRAINED_NEGATION)
        self.model = RobertaForTokenClassification.from_pretrained(
            **HF_FROM_PRETRAINED_NEGATION
        )

        super().__init__(**kwargs)

    @property
    def qualifier_factories(self) -> dict[str, QualifierFactory]:
        return {
            "Negation": QualifierFactory(
                "Negation", ["Affirmed", "Negated"], default="Affirmed"
            )
        }

    def _detect_qualifiers(self, doc: Doc):
        for ent in doc.ents:
            text, ent_start_char, ent_end_char = self._prepare_ent(ent)

            prob = self._predict(
                text,
                ent_start_char,
                ent_end_char,
                prob_indices=[0, 2],
                prob_aggregator=self.prob_aggregator,
            )

            if prob > self.negation_threshold:
                self.add_qualifier_to_ent(
                    ent,
                    self.qualifier_factories["Negation"].create("Negated", prob=prob),
                )


@Language.factory(
    name="clinlp_experiencer_transformer",
    requires=["doc.ents"],
    assigns=[f"span._.{ATTR_QUALIFIERS}"],
    default_config=_defaults_experiencer_transformer,
)
@clinlp_autocomponent
class ExperiencerTransformer(QualifierTransformer):
    def __init__(
        self,
        nlp: Language,
        other_threshold: float = _defaults_experiencer_transformer["other_threshold"],
        **kwargs,
    ) -> None:
        self.nlp = nlp
        self.other_threshold = other_threshold

        self.tokenizer = AutoTokenizer.from_pretrained(**HF_FROM_PRETRAINED_EXPERIENCER)
        self.model = RobertaForTokenClassification.from_pretrained(
            **HF_FROM_PRETRAINED_EXPERIENCER
        )

        super().__init__(**kwargs)

    @property
    def qualifier_factories(self) -> dict[str, QualifierFactory]:
        return {
            "Experiencer": QualifierFactory(
                "Experiencer", ["Patient", "Other"], default="Patient"
            )
        }

    def _detect_qualifiers(self, doc: Doc):
        for ent in doc.ents:
            text, ent_start_char, ent_end_char = self._prepare_ent(ent)

            prob = self._predict(
                text,
                ent_start_char,
                ent_end_char,
                prob_indices=[1, 3],
                prob_aggregator=self.prob_aggregator,
            )

            if prob > self.other_threshold:
                self.add_qualifier_to_ent(
                    ent,
                    self.qualifier_factories["Experiencer"].create("Other", prob=prob),
                )



import inspect
from inspect import Parameter, Signature

from makefun import with_signature


class _UnusedArgument:
    pass


def get_class_init_signature(cls):
    args = []
    kwargs = {}

    for mro_class in cls.__mro__:
        if "__init__" in mro_class.__dict__:
            argspec = inspect.getfullargspec(mro_class)

            if argspec.defaults is None:
                args += argspec.args[1:]
            else:
                first_arg_wit_default = len(argspec.args) - len(argspec.defaults)
                args += argspec.args[1:first_arg_wit_default]
                kwargs |= dict(
                    zip(argspec.args[first_arg_wit_default:], argspec.defaults)
                )

    return args, kwargs


def clinlp_autocomponent(cls):
    component_args, component_kwargs = get_class_init_signature(cls)

    params = []

    make_component_args = component_args.copy()

    if "nlp" not in make_component_args:
        make_component_args = ["nlp"] + make_component_args

    if "name" not in make_component_args:
        make_component_args = ["name"] + make_component_args

    for arg in make_component_args:
        params.append(
            Parameter(
                arg, kind=Parameter.POSITIONAL_OR_KEYWORD, default=_UnusedArgument()
            )
        )

    for kwarg, default in component_kwargs.items():
        params.append(
            Parameter(kwarg, kind=Parameter.POSITIONAL_OR_KEYWORD, default=default)
        )

    @with_signature(Signature(params), func_name="make_component")
    def make_component(*args, **kwargs):
        if len(args) > 0:
            raise RuntimeError("Please pass all arguments as keywords.")

        cls_kwargs = {
            k: v
            for k, v in kwargs.items()
            if (k in component_args or k in component_kwargs)
            and (not isinstance(v, _UnusedArgument))
        }

        return cls(**cls_kwargs)

    return make_component
